@@ -135,10 +135,12 @@ class Bolt:
         # Carry out Bolt subclass imports locally to avoid circular dependency issues.
         from neo4j.io._bolt3 import Bolt3
         from neo4j.io._bolt4x0 import Bolt4x0
+        from neo4j.io._bolt4x1 import Bolt4x1
 
         handlers = {
             Bolt3.PROTOCOL_VERSION: Bolt3,
-            Bolt4x0.PROTOCOL_VERSION: Bolt4x0
+            Bolt4x0.PROTOCOL_VERSION: Bolt4x0,
+            Bolt4x1.PROTOCOL_VERSION: Bolt4x1,
         }
 
         if protocol_version is None:
@@ -175,12 +177,13 @@ class Bolt:
             return protocol_version
 
     @classmethod
-    def open(cls, address, *, auth=None, timeout=None, **pool_config):
+    def open(cls, address, *, auth=None, timeout=None, routing_context=None, **pool_config):
         """ Open a new Bolt connection to a given server address.
 
         :param address:
         :param auth:
-        :param timeout: The connection timeout
+        :param timeout: the connection timeout in seconds
+        :param routing_context: dict containing routing context
         :param pool_config:
         :return:
         :raise BoltHandshakeError: raised if the Bolt Protocol can not negotiate a protocol version.
@@ -198,11 +201,15 @@ class Bolt:
         if pool_config.protocol_version == (3, 0):
             # Carry out Bolt subclass imports locally to avoid circular dependency issues.
             from neo4j.io._bolt3 import Bolt3
-            connection = Bolt3(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent)
+            connection = Bolt3(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
         elif pool_config.protocol_version == (4, 0):
             # Carry out Bolt subclass imports locally to avoid circular dependency issues.
             from neo4j.io._bolt4x0 import Bolt4x0
-            connection = Bolt4x0(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent)
+            connection = Bolt4x0(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
+        elif pool_config.protocol_version == (4, 1):
+            # Carry out Bolt subclass imports locally to avoid circular dependency issues.
+            from neo4j.io._bolt4x1 import Bolt4x1
+            connection = Bolt4x1(address, s, pool_config.max_connection_lifetime, auth=auth, user_agent=pool_config.user_agent, routing_context=routing_context)
         else:
             log.debug("[#%04X]  S: <CLOSE>", s.getpeername()[1])
             s.shutdown(SHUT_RDWR)
@@ -211,7 +218,14 @@ class Bolt:
             supported_versions = Bolt.protocol_handlers().keys()
             raise BoltHandshakeError("The Neo4J server does not support communication with this driver. This driver have support for Bolt Protocols {}".format(supported_versions), address=address, request_data=handshake, response_data=data)
 
-        connection.hello()
+        try:
+            connection.hello()
+        except Exception as error:
+            log.debug("[#%04X]  C: <CLOSE> %s", s.getsockname()[1], str(error))
+            s.shutdown(SHUT_RDWR)
+            s.close()
+            raise error
+
         return connection
 
     @property
@@ -493,27 +507,35 @@ class IOPool:
 class BoltPool(IOPool):
 
     @classmethod
-    def open(cls, address, *, auth, pool_config, workspace_config):
+    def open(cls, address, *, auth, pool_config, workspace_config, routing_context=None):
         """Create a new BoltPool
 
         :param address:
         :param auth:
         :param pool_config:
         :param workspace_config:
+        :param routing_context:
         :return: BoltPool
         """
 
-        def opener(addr, timeout):
-            return Bolt.open(addr, auth=auth, timeout=timeout, **pool_config)
+        if routing_context is None:
+            routing_context = {}
+        elif "address" in routing_context:
+            raise ConfigurationError("The key 'address' is reserved for routing context.")
+        routing_context["address"] = str(address)
 
-        pool = cls(opener, pool_config, workspace_config, address)
+        def opener(addr, timeout):
+            return Bolt.open(addr, auth=auth, timeout=timeout, routing_context=routing_context, **pool_config)
+
+        pool = cls(opener, pool_config, workspace_config, routing_context, address)
         seeds = [pool.acquire() for _ in range(pool_config.init_size)]
         pool.release(*seeds)
         return pool
 
-    def __init__(self, opener, pool_config, workspace_config, address):
+    def __init__(self, opener, pool_config, workspace_config, routing_context, address):
         super(BoltPool, self).__init__(opener, pool_config, workspace_config)
         self.address = address
+        self.routing_context = routing_context
 
     def __repr__(self):
         return "<{} address={!r}>".format(self.__class__.__name__, self.address)
@@ -539,10 +561,17 @@ class Neo4jPool(IOPool):
         :return: Neo4jPool
         """
 
-        def opener(addr, timeout):
-            return Bolt.open(addr, auth=auth, timeout=timeout, **pool_config)
+        address = addresses[0]
+        if routing_context is None:
+            routing_context = {}
+        elif "address" in routing_context:
+            raise ConfigurationError("The key 'address' is reserved for routing context.")
+        routing_context["address"] = str(address)
 
-        pool = cls(opener, pool_config, workspace_config, routing_context, addresses)
+        def opener(addr, timeout):
+            return Bolt.open(addr, auth=auth, timeout=timeout, routing_context=routing_context, **pool_config)
+
+        pool = cls(opener, pool_config, workspace_config, routing_context, address)
 
         try:
             pool.update_routing_table(database=workspace_config.database)
@@ -552,7 +581,7 @@ class Neo4jPool(IOPool):
         else:
             return pool
 
-    def __init__(self, opener, pool_config, workspace_config, routing_context, addresses):
+    def __init__(self, opener, pool_config, workspace_config, routing_context, address):
         """
 
         :param opener:
@@ -563,15 +592,10 @@ class Neo4jPool(IOPool):
         """
         super(Neo4jPool, self).__init__(opener, pool_config, workspace_config)
         # Each database have a routing table, the default database is a special case.
-        log.debug("[#0000]  C: <NEO4J POOL> routing addresses %r", addresses)
-        self.init_address = addresses[0]
-        self.routing_tables = {workspace_config.database: RoutingTable(database=workspace_config.database, routers=addresses)}
+        log.debug("[#0000]  C: <NEO4J POOL> routing address %r", address)
+        self.address = address
+        self.routing_tables = {workspace_config.database: RoutingTable(database=workspace_config.database, routers=[address])}
         self.routing_context = routing_context
-        if self.routing_context is None:
-            self.routing_context = {}
-        elif "address" in self.routing_context:
-            raise ConfigurationError("The key 'address' is reserved for routing context.")
-        self.routing_context["address"] = str(self.init_address)
         self.refresh_lock = Lock()
 
     def __repr__(self):
@@ -615,7 +639,7 @@ class Neo4jPool(IOPool):
         :param address: router address
         :param timeout: seconds
         :param database: the data base name to get routing table for
-        :param init_address: the address by which the client initially contacted the server as a hint for inclusion in the returned routing table.
+        :param address: the address by which the client initially contacted the server as a hint for inclusion in the returned routing table.
 
         :return: list of routing records or
                  None if no connection could be established
@@ -650,6 +674,7 @@ class Neo4jPool(IOPool):
         # Carry out Bolt subclass imports locally to avoid circular dependency issues.
         from neo4j.io._bolt3 import Bolt3
         from neo4j.io._bolt4x0 import Bolt4x0
+        from neo4j.io._bolt4x1 import Bolt4x1
 
         from neo4j.api import (
             SYSTEM_DATABASE,
@@ -668,43 +693,12 @@ class Neo4jPool(IOPool):
 
         try:
             with self._acquire(address, timeout) as cx:
-                _, _, server_version = (cx.server_info.agent or "").partition("/")
                 log.debug("[#%04X]  C: <ROUTING> query=%r", cx.local_port, self.routing_context or {})
 
                 if database is None:
                     database = self.workspace_config.database
 
-                # TODO: This logic should be inside the Bolt subclasses, because it can change depending on Bolt Protocol Version.
-                if cx.PROTOCOL_VERSION == Bolt3.PROTOCOL_VERSION:
-                    if database != DEFAULT_DATABASE:
-                        raise ConfigurationError("Database name parameter for selecting database is not supported in Bolt Protocol {!r}. Database name {!r}. Server Agent {!r}.".format(
-                                Bolt3.PROTOCOL_VERSION, database, cx.server_info.agent))
-                    cx.run(
-                        "CALL dbms.cluster.routing.getRoutingTable($context)",  # This is an internal procedure call. Only available if the Neo4j 3.5 is setup with clustering.
-                        {"context": self.routing_context},
-                        mode="r",  # Bolt Protocol Version(3, 0) supports mode
-                        on_success=metadata.update,
-                        on_failure=fail,
-                    )
-                elif cx.PROTOCOL_VERSION == Bolt4x0.PROTOCOL_VERSION:
-                    if database == DEFAULT_DATABASE:
-                        cx.run(
-                            "CALL dbms.routing.getRoutingTable($context)",
-                            {"context": self.routing_context},
-                            mode="r",
-                            db=SYSTEM_DATABASE,
-                            on_success=metadata.update,
-                            on_failure=fail,
-                        )
-                    else:
-                        cx.run(
-                            "CALL dbms.routing.getRoutingTable($context, $database)",
-                            {"context": self.routing_context, "database": database},
-                            mode="r",
-                            db=SYSTEM_DATABASE,
-                            on_success=metadata.update,
-                            on_failure=fail,
-                        )
+                cx.run_get_routing_table(on_success=metadata.update, on_failure=fail, database=database)
                 cx.pull(on_success=metadata.update, on_records=records.extend)
                 cx.send_all()
                 cx.fetch_all()
@@ -1012,7 +1006,7 @@ def _handshake(s, resolved_address):
         # response, the server has closed the connection
         log.debug("[#%04X]  S: <CLOSE>", local_port)
         s.close()
-        raise BoltHandshakeError("Connection to {address} closed without handshake response".format(address=resolved_address), address=resolved_address, request_data=handshake, response_data=None)
+        raise ServiceUnavailable("Connection to {address} closed without handshake response".format(address=resolved_address))
     if data_size != 4:
         # Some garbled data has been received
         log.debug("[#%04X]  S: @*#!", local_port)
